%% tcpl__connexity.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__connexity.pl,v 1.1 2002/10/17 15:30:17 coquery Exp $

%% Handle the order in which the clauses must be analysed

:- module(tclp__connexity, [build_ordered_connex_part_list/2]).

%:- use_module(tclp(tclp__def)).
:- use_module(library(lists)).
:- use_module(library(ugraphs)).
:- use_module(tclp(tclp__handle_goals)).

%% We use the call graph to compute connex parts
%% The nodes are of the form p/n

%% building the call graph

%% takes all clauses and build the list of edges corresponding to the call
%% graph
build_edges(Phrases, Predicates, Edges) :-
        build_edges(Phrases, Predicates, [], Edges).

build_edges([], _, Edges, Edges).
build_edges([ PhraseLoc | Rest ], Predicates, Accu, Edges) :-
        build_edges1(PhraseLoc, Predicates, Accu, Inter),
        build_edges(Rest, Predicates, Inter, Edges).

%% build the edges for one phrase
%build_edges1((Head :- Body), Predicates,CurrentEdges, ResultEdges) :- !,
%        functor(Head,P,N),
%        memberchk(P/N, Predicates),
%        build_edges2b(Head, N, P/N, Predicates, CurrentEdges, InterEdges),
%        build_edges2(Body, P/N, Predicates,InterEdges, ResultEdges).
%build_edges1((:- _),_,CurrentEdges,CurrentEdges) :- !.
%build_edges1(Head,Predicates,CurrentEdges,ResultEdges) :-
%        functor(Head,P,N),
%        memberchk(P/N, Predicates),
%        build_edges2b(Head, N, P/N, Predicates, CurrentEdges, ResultEdges).
build_edges1(Phrase-Loc, Predicates,CurrentEdges, ResultEdges) :-
        tclp__clause(Phrase,Loc,Heads,Bodies)
        ->
        build_edges3(Heads,Heads,CurrentEdges,InterEdges),
        build_edges1b(Heads,Bodies,Predicates, InterEdges, ResultEdges)
        ;
        CurrentEdges = ResultEdges.
build_edges1b([],_,_,CurrentEdges,CurrentEdges).
build_edges1b([Head-_-_|HDs],Bodies,Predicates,CurrentEdges, ResultEdges) :-
        functor(Head,P,N),
        memberchk(P/N, Predicates),
        build_edges2b(Head, N, P/N, Predicates, CurrentEdges, InterEdges),
        build_edges1c(P/N,Bodies,Predicates,InterEdges,InterEdges2),
        build_edges1b(HDs,Bodies,Predicates,InterEdges2, ResultEdges).
build_edges1c(_,[],_,CurrentEdges,CurrentEdges).
build_edges1c(P/N,[Body-_|Bodies],Predicates,CurrentEdges, ResultEdges) :-
        build_edges2(Body, P/N, Predicates,CurrentEdges, InterEdges),
        build_edges1c(P/N, Bodies, Predicates,InterEdges,ResultEdges).

%% build the edges for one term, given a starting point
build_edges2(X,_,_,CurrentEdges, CurrentEdges) :- var(X),!.
build_edges2(Term,P/N,Predicates, CurrentEdges, ResultEdges) :-
        functor(Term,P1,N1),
        (
          memberchk(P1/N1,Predicates)
        ->
          InterEdges = [ (P/N)-(P1/N1) | CurrentEdges ]
        ;
          InterEdges = CurrentEdges
        ),
        build_edges2b(Term, N1, P/N, Predicates, InterEdges, ResultEdges).

%% build the edges for the subterms of the given term
build_edges2b(_, 0, _, _, CurrentEdges, CurrentEdges) :- !.
build_edges2b(Term, I, P/N, Predicates, CurrentEdges, ResultEdges) :-
        arg(I,Term,Arg),
        build_edges2(Arg,P/N,Predicates,CurrentEdges,InterEdges),
        I1 is I-1,
        build_edges2b(Term,I1,P/N, Predicates, InterEdges, ResultEdges).


%% builds the edges that link all heads together
build_edges3([],_,CurrentEdges,CurrentEdges).
build_edges3([Head-_-_|HDs], AllHDs, CurrentEdges, ResultEdges) :-
        build_edges3b(Head, AllHDs, CurrentEdges, InterEdges),
        build_edges3(HDs, AllHDs, InterEdges, ResultEdges).
build_edges3b(_,[],CurrentEdges,CurrentEdges).
build_edges3b(Head,[Head-_-_|AllHDs],CurrentEdges,ResultEdges) :- !,
        build_edges3b(Head,AllHDs,CurrentEdges,ResultEdges).
build_edges3b(Head,[OtherHead-_-_|AllHDs],CurrentEdges,ResultEdges) :-
        functor(Head,P1,N1), functor(OtherHead,P2,N2),
        InterEdges = [ (P1/N1)-(P2/N2) | CurrentEdges ],
        build_edges3b(Head,AllHDs,InterEdges,ResultEdges).

%% Extracts all the defined predicates from the phrases
build_vertices(Phrases, Vertices) :-
        build_vertices2(Phrases, UnsortedVertices),
        sort(UnsortedVertices, Vertices).
build_vertices2([],[]).
%build_vertices2([ (Head :- _)-_ | Rest ],[ P/N | Vertices]) :- !,
%        functor(Head,P,N),
%        build_vertices2(Rest,Vertices).
%build_vertices2([ (:- _)-_ | Rest],Vertices) :- !,
%        build_vertices2(Rest,Vertices).
%build_vertices2([ Head-_ | Rest ],[ P/N | Vertices]) :- !,
%        functor(Head,P,N),
%        build_vertices2(Rest,Vertices).
build_vertices2([Phrase-Loc | Rest ] , Result) :-
        (
          tclp__clause(Phrase,Loc,Heads,_)
        ->
          build_vertices2b(Heads,HSpecs),
          append(HSpecs,Next,Result)
        ;
          Next=Result
        ),
        build_vertices2(Rest,Next).
build_vertices2b([],[]).
build_vertices2b([Head-_-_|HDs],[P/N|Next]) :-
        functor(Head,P,N),
        build_vertices2b(HDs,Next).

%% builds the call graph from a list of phrases
build_graph(Phrases, Vertices, Graph) :-
        build_vertices(Phrases, Vertices),
        build_edges(Phrases, Vertices, Edges),
        vertices_edges_to_ugraph(Vertices, Edges, Graph).


%% builds a list which is the reverse suffix deep course of the graph
reverse_suffix_deep_course(Graph, Vertices, Result) :-
        reverse_suffix_deep_course(Graph, Vertices, [], _, [], Result).

reverse_suffix_deep_course(_, [], Visited, Visited, Result, Result).
reverse_suffix_deep_course(Graph, [ Vertex | Rest ],
                           Visited,Visited3,
                           Current, Result) :-
        reverse_suffix_deep_course2(Graph, Vertex,
                                    Visited, Visited2,
                                    Current, Inter),
        reverse_suffix_deep_course(Graph, Rest,
                                   Visited2, Visited3,
                                   Inter, Result).

reverse_suffix_deep_course2(_, Vertex,
                            Visited, Visited,
                            Current, Current) :-
        memberchk(Vertex, Visited),!.
reverse_suffix_deep_course2(Graph, Vertex,
                            Visited, Visited2,
                            Current, [Vertex|Inter]) :-
        neighbours(Vertex, Graph, Neighbours),
        reverse_suffix_deep_course(Graph, Neighbours,
                                   [ Vertex | Visited], Visited2,
                                   Current, Inter).

%% makes a deep course in the graph (with only one visit per vertex)
%% builds a list of list of vertex corresponding to the course
deep_course(Graph, Vertices, SubGraphs) :-
        deep_course(Graph, Vertices, [], SubGraphs).

deep_course(_Graph, [], SubGraphs, SubGraphs).
deep_course(Graph, [ Vertex | Rest ], Current, Result) :-
        deep_course2(Graph, Vertex, Current, Inter),
        deep_course(Graph, Rest, Inter, Result).

deep_course2(_Graph, Vertex, Current, Current) :-
        member(SubGraph, Current),
        member(Vertex, SubGraph),!.
deep_course2(Graph, Vertex, Current, [ SubResult | Current ]) :-
        deep_course3(Graph, Vertex, Current, [], SubResult).

deep_course3(_Graph, Vertex, Visited, Result, Result) :-
        member(SubGraph, Visited),
        member(Vertex, SubGraph),!.
deep_course3(_Graph, Vertex, _, Result, Result) :-
        member(Vertex, Result),!.
deep_course3(Graph, Vertex, Visited, Current, Result) :-
        neighbours(Vertex, Graph, Neighbours),
        deep_course4(Graph, Neighbours, Visited, 
                     [ Vertex | Current ], Result).

deep_course4(_Graph, [], _, Result, Result).
deep_course4(Graph, [Vertex|Rest], Visited, Current, Result) :-
        deep_course3(Graph, Vertex, Visited, Current, Inter),
        deep_course4(Graph, Rest, Visited, Inter, Result).

%% makes a list of empty lists based with the same length of the given list
%% of lists + 1
build_empty_lists_list([],[[]]).
build_empty_lists_list([_|Rest],[[]|Rest2]) :-
        build_empty_lists_list(Rest, Rest2).

%% adds a phrase to a phrase partition, given its corresponding functor and
%% a predicate partition
add_to_partition(_,_,_,[],_) :- !, fail.
add_to_partition(Phrase,_,[],[List],[[Phrase|List]]).
add_to_partition(Phrase,Id,
                 [ SomeIds | _ ],
                 [ SomePhrases | RestPhrases ],
                 [ [ Phrase | SomePhrases ] | RestPhrases ]) :-
        member(Id, SomeIds),!.
add_to_partition(Phrase,Id,
                 [_|OtherIds],
                 [SomePhrases|RestPhrases],
                 [SomePhrases|RestPhrases2]) :-
        add_to_partition(Phrase,Id, OtherIds, RestPhrases, RestPhrases2).

%% given an ordered vertices partition and a list of phrases,
%% returns an ordered phrases partition
build_connex_parts(Phrases, VertexPartition, PhrasePartition) :-
        build_empty_lists_list(VertexPartition, Start),
        build_connex_parts(Phrases, VertexPartition, Start, PhrasePartition).

build_connex_parts([], _, Start, Start).
build_connex_parts([Phrase|Rest], VertexPartition, Start, PhrasePartition) :-
        add_phrase_to_partition(Phrase, VertexPartition,
                                Start, Inter),
        build_connex_parts(Rest, VertexPartition, Inter, PhrasePartition).

%add_phrase_to_partition(Phrase-TermData, VertexPartition, Start, Result) :-
%        (
%          Phrase = (Head :- _)
%        ->
%          functor(Head,P,N), Id = P/N
%        ;
%          Phrase = (:- _)
%        ->
%          Id = '$directive'
%        ;
%          functor(Phrase,P,N), Id = P/N
%        ),
%        add_to_partition(Phrase-TermData, Id, VertexPartition, Start, Result).
add_phrase_to_partition(Phrase-TermData, VertexPartition, Start, Result) :-
        (
          tclp__clause(Phrase,TermData,[Head-_-_|_],_)
        ->
          functor(Head,P,N), Id = P/N
        ;
          Id = '$directive'
        ),
        add_to_partition(Phrase-TermData, Id, VertexPartition, Start, Result).

%% builds an ordered list of list of phrases made using the order given by
%% the clauses call graph
build_ordered_connex_part_list(Phrases, List) :-
        build_graph(Phrases, Vertices, Graph),
        reverse_suffix_deep_course(Graph, Vertices, OrderedVertices),
        transpose(Graph, IGraph),
        deep_course(IGraph, OrderedVertices, OrderedVerticesPartition),
        build_connex_parts(Phrases, OrderedVerticesPartition, List).
